/* Faktorisierungsprogramm
   written 03/1998 by Thorsten Reinecke
   
   Fr den mathematischen Hintergrund siehe:
   Riesel, Hans: "Prime Numbers and computer methods for factorization",
   Boston; Basel; Stuttgart: Birkhuser, 1985
    
   Das Programm setzt einige der dort beschriebenen Algorithmen (in freier
   Variation) um.
   Zunchst findet eine Probedivision fr kleine Primzahlen statt.
   Anschlieend wird mit der Pollard-rho-Methode nach weiteren kleineren
   Faktoren gesucht.
   Falls um eins verminderte Faktoren ausschlielich in kleinere Teiler
   zerfallen, so werden diese mittels der Pollard-Phi-Methode entdeckt.
   
   Wenn die bis dahin verbleibende Zahl nachweislich zusammengesetzt
   (also nicht pseudoprim) ist, wird eine Faktorisierung durch
   das Quadratische Sieb versucht.

   Das lineare Gleichungssystem wird dabei simulatan mit den eingefgten
   Relationen gelst.

   In seiner jetzigen Fassung sollte man mit dem Programm
   Zahlen bis um 70 Dezimalstellen in ertrglicher Zeit faktorisieren knnen.
   
   Es fehlt allerdings noch eine ganze Reihe an Optimierungen und Algorithmen.
   So fehlt beispielsweise eine Implementierung eliptischer Kurven zur
   Ermittlung von Faktoren "gemigter Gre". Dies wre das einzige
   Verfahren, mit dem sich momentan auch bei sehr groen Zahlen nichttriviale
   Faktoren entdecken lassen (da seine Laufzeit von der Gre des entdeckten
   Faktors abhngig ist).
   Fr ein effizienteres Sieben andererseits wre insbesondere die
   Implementierung des "Multiple Polynomial Quadratic Sieve" oder gar des
   "Number-Field-Sieve" sinnvoll. (Siebverfahren faktorisieren grundstzlich
   in Abhngigkeit von der Gre der zu faktorisierenden Zahl; sie sind
   besonders effizient, wenn die Zahl lediglich zwei Primteiler hat und
   diese von der Stelligkeit her nicht allzuweit voneinander entfernt liegen.)

   Auerdem wre dann das komplette Programm anzupassen und zu optimieren,
   insbesondere auch die lineare Algebra, die man vielleicht auf den
   Hintergrundspeicher auslagern knnte.
   (Auch so ist das Programm schon hufiger verndert worden, so da
   ich es nochmal "sauberschreiben" mte...)
   Mal sehen, wann ich Zeit und Lust dazu habe... ;)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gmp.h>
#include <BitString.h>
#include <list>
#include <set>
#include <vector>
#include "qsieve.h"

#define FALSE 0
#define TRUE 1

mpz_t n, /* zu faktorisierende Zahl */
      kN, /* das mit einem gnstigen Vorfaktor multiplizierte n
	     (nachdem n um einfach zu findende Faktoren reduziert wurde) */
      r, /* Quadratwurzel von n (oder bei optimierter Fassung k*n) */
      x,y; /* globale Hilfsvariablen,
	      jede Unterfunktion kann sie nutzen, mu aber davon ausgehen,
	      da sie jederzeit von anderen Funktionen verndert werden! */
	      

#define Primbasis_size 20000 /* Maximale Gre der Primzahlbasis */ 
/* Hinweis: Je grer die zu faktorisierende Zahl ist, desto grer
   mte prinzipiell auch die Primzahlbasis sein.
   Aber: Das lineare Gleichungssystem sprengt bei groer Primzahlbasis
   den Speicher!
   Fr kleinere zu faktorisierende Zahlen wirkt sich dagegen eine grere
   Primzahlbasis verlangsamend aus. Aber eine dynamische Anpassung ist noch
   nicht implementiert.
   Sinnvolle Werte liegen zwischen 1000 und 50000 (je nach verfgbarem
   Speicher).
*/

int Primzahlen[Primbasis_size]; /* fr eine effiziente Probedivision
				 alle Primzahlen der Faktorbasis hier
				 akkumulieren */

int Primzahl    = 1;
int Primzahl_nr = 1;
int Delta_der_Primzahl[2]; /* Es gibt zwei Wurzeln, also zwei Deltas */
int Delta_der_Primzahlen[Primbasis_size][2]; /* fr den effizienten Zugriff
						werden die einmal berechneten
						Werte hier abgelegt */

#define Factor_Threshold 1.1 /* Multiplikator, der grere Faktoren ber die
				Faktorbasis hinaus erlaubt.
				Sinnvolle Werte:
				1.0 bis grter Faktor der Faktorbasis */


#define MAX_Siebgroesse 2000000
/* In Intervallen von MAX_Siebgroesse werden die Relationen gesiebt.
   Hinweis: Je grer MAX_Siebgroesse, desto effizienter geht das Aussieben
   der Relationen. Gleichzeitig erhht sich aber auch der Speicherbedarf.
   Der Einflu der Siebgre ist allerdings geringer als der der gewhlten
   Primzahlbasis. Ein Wert zwischen 1 Mio und 5 Mio drfte daher vollkommen
   ausreichen.
*/

int Siebgroesse = MAX_Siebgroesse;

typedef short int TSiebElement;
typedef vector<TSiebElement> TSieb; /* Quadratisches Sieb */
TSieb *Sieb;
#define log_Multiplikator 100 /* Siebwert wird nach damit
				 nach short int konvertiert */

mpz_t Sieboffset; /* Initialwert im Programm: -MAX_Siebgroesse/2 */
/* Es wird jeweils im Intervall 
   [Sieboffset .. Sieboffset+Siebgroesse] gesiebt! */

int gefundene_Relationen = 0; /* Anzahl der bereits gefundenen Relationen */

int biggest_Prime_in_Factorbase;

mpz_t Deltas_der_Relationen [(int)(Factor_Threshold*Primbasis_size)]; 
/* hier werden die Deltawerte der gefundenen Relationen gespeichert */


void pollard(int runden) /* Pollard-rho-Methode */
{
  cout << "Trying Pollard-rho up to " << runden << endl;
  mpz_t a,a2;
  mpz_init(a); mpz_init(a2);
  mpz_set_ui(a,1); mpz_set(a2,a);
  int i=0;
  do
    {
      mpz_mul(a,a,a); mpz_add_ui(a,a,1); mpz_mod(a,a,n);
      mpz_mul(a2,a2,a2); mpz_add_ui(a2,a2,1); mpz_mod(a2,a2,n);
      mpz_mul(a2,a2,a2); mpz_add_ui(a2,a2,1); mpz_mod(a2,a2,n);
      mpz_sub(x,a2,a);
      mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)
	{
	  do
	    {
	      mpz_out_str(stdout,10,x);
	      if (mpz_probab_prime_p(x,20))
		printf(" (%i) ist Faktor.\n",mpz_sizeinbase(x,10));
	      else
		printf(" (%i) ist zusammengesetzter Faktor.\n",
		       mpz_sizeinbase(x,10));
	      mpz_div(n,n,x);
	      mpz_gcd(x,x,n);
	    } while (mpz_cmp_ui(x,1)!=0);
	  if (mpz_probab_prime_p(n,20)) break; /* Rest wahrscheinlich prim */
	}
    } while (++i<runden);
  mpz_clear(a); mpz_clear(a2);
}


int pollard_reduce (int runden)
{
  /* y um alle Faktoren verkleinern, die innerhalb von "runden" Pollardrunden
     entdeckt werden... */

  mpz_t a,a2,x_prod;
  int i=0;
  mpz_init(a); mpz_init(a2); mpz_init(x_prod);
  mpz_set_ui(a,1); mpz_set(a2,a); mpz_set_ui(x_prod,1);
  if (mpz_sgn(y)<0) mpz_neg(y,y); // y = abs(y)  
  do
    {
      mpz_mul(a,a,a); mpz_add_ui(a,a,1); mpz_mod(a,a,y);
      mpz_mul(a2,a2,a2); mpz_add_ui(a2,a2,1); mpz_mod(a2,a2,y);
      mpz_mul(a2,a2,a2); mpz_add_ui(a2,a2,1); mpz_mod(a2,a2,y);
      mpz_sub(x,a2,a);
      mpz_mul(x_prod,x_prod,x); mpz_mod(x_prod,x_prod,y);
      mpz_gcd(x,x_prod,y); mpz_div(y,y,x);
    } while (++i<runden);
  mpz_clear(a); mpz_clear(a2); mpz_clear(x_prod);
  return (mpz_cmp_ui(y,1)==0);
}


int is_prime(int p)
{
 int t = 2;
 while (t*t<=p) 
   {
     if ((p%t)==0) return FALSE;
     t++;
   }
 return TRUE;
}


void pollard_phi(int phase1, int phase2)
  /* Pollard-Phi Methode in zwei Phasen */
{
  mpz_t a,a_old,b,x,h;
  mpz_init(a); mpz_init(a_old); mpz_init(b); mpz_init(x); mpz_init(h);
  do
    {
      cout << "Pollard-Phi Phase 1" << endl;
      mpz_set_ui(a,2);
      mpz_set(a_old,a); /* Wiederaufnahmewert, falls mehere Faktoren zusammen
			   entdeckt werden */
      for (int i=phase1/2; i<phase1; )
	{
	  mpz_set(a_old,a); /* Wiederaufnahmewert, falls mehrere Faktoren
			       zusammen entdeckt werden */
	  for (int j=0; j++<5000; i++ ) mpz_powm_ui(a,a,i,n);
	  
	  {
	    printf("%i/%i\r",i,phase1); fflush(stdout);
	    mpz_sub_ui(x,a,1); mpz_gcd(x,x,n);
	    if (mpz_cmp_ui(x,1)!=0) 
	      { 
		if (!mpz_probab_prime_p(x,10))
		  { /* versuchen, den entdeckten zusammengesetzen Faktor
		       aufzuspalten */
		    cout << "trying to split composed Factor..." << endl;
		    i-=5000;
		    mpz_set(a,a_old);
		    do
		      {
			mpz_powm_ui(a,a,i,n);
			mpz_sub_ui(x,a,1); mpz_gcd(x,x,n);
			i++;
		      } while (mpz_cmp_ui(x,1)==0);
		  }
		break;
	      }	
	    mpz_set(a_old,a);
	  }
	}
      
      mpz_sub_ui(x,a,1); mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)==0) // kein Teiler gefunden -> Phase 2 starten
	{
	  cout << "Pollard-Phi Phase 2" << endl;
	  mpz_powm_ui(b,a,2,n);
	  {
	    int i;
	    for (i=phase1; !is_prime(i); i--);
	    mpz_powm_ui(a,a,i,n);
	  }
	  mpz_set(a_old,a);
	  for (int i=1; i<=phase2; i++)
	    {
	      printf("%i/%i\r",i,phase2); fflush(stdout);
	      for (int j=0; j<phase1; j++)
		{
		  mpz_mul(a,a,b); mpz_mod(a,a,n);
		  mpz_sub_ui(h,a,1); mpz_mul(x,x,h); mpz_mod(x,x,n);
		}
	      mpz_gcd(x,x,n); 
	      if (mpz_cmp_ui(x,1)!=0) 
		{
		  if (!mpz_probab_prime_p(x,10))
		    { /* versuchen, den entdeckten zusammengesetzen Faktor
			 aufzuspalten */
		      cout << "trying to split composed Factor..." << endl;
		      mpz_set(a,a_old); // Wiederaufnahmewert
		      do
			{
			  mpz_mul(a,a,b); mpz_mod(a,a,n);
			  mpz_sub_ui(h,a,1); mpz_gcd(x,h,n);
			} while (mpz_cmp_ui(x,1)!=1);
		    }
		  break;
		}
	      mpz_set(a_old,a);
	    }
	}
      
      if (mpz_cmp(x,n)==0) // ist Teiler ist trivial (d.h. n/n=1)?
	{
	  cout << "no success.." << endl;
	  break; // Teiler ist trivial (d.h. n/n=1)
	}
      if (mpz_cmp_ui(x,1)!=0)  /* Teiler gefunden? */
	{
	  mpz_out_str(stdout,10,x);
	  if (mpz_probab_prime_p(x,20))
	    printf(" (%i) ist Faktor.\n",mpz_sizeinbase(x,10));
	  else
	    printf(" (%i) ist zusammengesetzter Faktor.\n",
		   mpz_sizeinbase(x,10));
	  mpz_div(n,n,x);
	  if (mpz_probab_prime_p(n,10)) break;
	}
    } while (mpz_cmp_ui(x,1)!=0);
  
  mpz_clear(a); mpz_clear(a_old); mpz_clear(b); mpz_clear(x); mpz_clear(h);
}

void fermat_like_methode(unsigned int bis)
{
  /* Fermat-hnliche Methode: wenn t die Zahl n teilt
     und abs(t - n/t) klein ist, so kann hiermit ein Teiler gefunden werden.
     Alternativ ist auch ein beliebiger Vorfaktor fr abs(k*t - n/t)
     realisiert.
     Diese Methode ist zwar nicht besonders effizient, da sie eher einer
     Probedivision entspricht; sie hilft aber, solche Zahlen besonders
     schnell zu faktorisieren, die auf "allzu naive" Weise als
     "besonders schwer" zu faktorisierende Zahlen konstruiert sind.
  */
  cout << "starte Fermat-hnliche Methode" << endl;

  mpz_t c,r,h,h2;
  mpz_init(c); mpz_init(r); mpz_init(h); mpz_init(h2);

  unsigned int delta;

  for (unsigned int k=1; k<=bis; k++)
    {
      cout << "Vorfaktor ist " << k << "   " << endl;
      mpz_mul_ui(c,n,k);
      mpz_mul_ui(c,c,4); // 4c ist notwendig fr "Quadrat-Formel"
      
      mpz_sqrt(r,c);
      for(delta=0; delta<1000000; delta++)
	{
	  if (delta%100000==0)
	    { printf("delta=%i\r",delta); fflush(stdout); }
	  mpz_add_ui(h,r,delta); mpz_mul(h,h,h); mpz_sub(h,h,c);
	  if (mpz_perfect_square_p(h))
	    {
	      { printf("delta=%i\n",delta); fflush(stdout); }
	      cout << "Square detected..." << endl;
	      mpz_sqrt(h2,h); mpz_add_ui(h,r,delta); mpz_sub(h,h,h2);
	      mpz_gcd(h,h,n);
	      if (mpz_cmp_ui(h,1)!=0) 
		{
		  mpz_out_str(stdout,10,h);
		  if (mpz_probab_prime_p(h,20))
		    printf(" (%i) ist Faktor.\n",mpz_sizeinbase(h,10));
		  else
		    printf(" (%i) ist zusammengesetzter Faktor.\n",
			   mpz_sizeinbase(h,10));
		  mpz_div(n,n,h);
		  fermat_like_methode(bis); // und noch einmal versuchen...
		  k=bis+1; break;
		}
	    }
	}
    }
  mpz_clear(h); mpz_clear(h2); mpz_clear(r);
  mpz_clear(c);
}



void easy_factor()
{
  /* einfache Faktorisierung versuchen:
     vielleicht kann die Zahl dadurch komplett faktorisiert werden
     oder zumindest um einige Faktoren reduziert werden... */

  if (mpz_sizeinbase(n, 10)>40)
    fermat_like_methode(1);
  /* Fermat-hnliche Faktorisierung versuchen, fr den Fall da die Zahl
     z.B. die Form (a-d)*(a+d) mit kleinem d hat... */
  
  cout << "Trialdivision up to 5000" << endl;
  /* Probedivision mit 2 und 3 */
  while (mpz_div_ui(y,n,2)==0) { mpz_set(n,y); printf("2 ist Faktor.\n"); }
  while (mpz_div_ui(y,n,3)==0) { mpz_set(n,y); printf("3 ist Faktor.\n"); }
  
  { /* Probedivision bis 5000 */
    unsigned long int p=5;
    while (p<5000)
      {
	while (mpz_div_ui(y,n,p)==0)
	  { mpz_set(n,y); cout << p << " ist Faktor." << endl; }
	p+=2;
	while (mpz_div_ui(y,n,p)==0)
	  { mpz_set(n,y); cout << p << " ist Faktor." << endl; }
	p+=4;
      } 
  }
  
  if (mpz_cmp_ui(n,1)==0) exit(0); /* 1 als Faktor: komplett faktorisiert */  
  
  if (!mpz_probab_prime_p(n,25))
    pollard(100000); /* Faktorisierung nach Pollard-Rho-methode versuchen */
  
  if (!mpz_probab_prime_p(n,25))
    pollard_phi(100000,50); /* Faktorisierung nach Pollard-Phi-Methode
			       versuchen */
  
  if (!mpz_probab_prime_p(n,25) && mpz_sizeinbase(n, 10)>40)
    fermat_like_methode(5);  /* Fermat-hnliche Faktorisierung versuchen */
  
  printf("Verbleibende Zahl:\n");
  mpz_out_str(stdout,10,n); printf(" (%i)\n",mpz_sizeinbase(n,10));
  
  if (mpz_cmp_ui(n,1)==0) exit(0); /* 1 als Faktor: komplett faktorisiert */  
  
  if (mpz_probab_prime_p(n,25))
    {
      printf("Verbleibende Zahl ist wahrscheinlich prim!\n");
      exit(0);
    }
}



unsigned int lucas_p, lucas_q, lucas_p_inv;
#define lucas_cache_size 250
int lucas_cache [lucas_cache_size][2];
int lucas_cache_index = 0;


int lucasv(int m)
{
  if (m==0) return 2;
  if (m==1) return lucas_p;
  
  /* Wert schonmal vorher berechnet? */
  for (int i=0; i<lucas_cache_index; i++)
    {
      if (lucas_cache[i][0]==m) return lucas_cache[i][1];
    }
  
  int wert; /* Rckgabewert fr lucas_v */
  // printf("lucasv(%i)\n",m);
  
  if (m%2==0) 
    {
      mpz_t h1,h2,p;
      
      mpz_init(h1); mpz_init(h2); mpz_init(p);
      mpz_set_si(h1,lucasv(m/2)); mpz_mul(h1,h1,h1); 
      
      mpz_set_ui(p,Primzahl);
      mpz_set_ui(h2,lucas_q); mpz_powm_ui(h2,h2,m/2,p);
      mpz_mul_ui(h2,h2,2);
      
      mpz_sub(h1,h1,h2);
      wert=mpz_mod_ui(h1,h1,Primzahl);
      
      mpz_clear(h1); mpz_clear(h2); mpz_clear(p);
    }
  else
    {
      mpz_t h1,h2;
      mpz_init(h1); mpz_init(h2);
      
      mpz_set_si(h1,lucasv(m+1));
      mpz_set_si(h2,lucasv(m-1));
      mpz_mul_ui(h2,h2,lucas_q);
      mpz_add(h1,h1,h2); 
      mpz_mul_ui(h1,h1,lucas_p_inv);
      wert=mpz_mod_ui(h1,h1,Primzahl);
      
      mpz_clear(h1); mpz_clear(h2);
    }
  
  /* Wert in den Cache schreiben... */
  if (lucas_cache_index<lucas_cache_size)
    {
      lucas_cache[lucas_cache_index][0]=m;
      lucas_cache[lucas_cache_index++][1]=wert;
    } else cerr << "Lucas-Cache sollte vergrert werden!" << endl;
  return wert;
}


int lucas()
{
  int v;
  if (Primzahl%4!=1)
    { printf("Fehler in Lucassequenz %i\n",Primzahl%4); exit(1); }

  /* lucas_p ermitteln */
  lucas_q=mpz_mod_ui(y,kN,Primzahl);
  lucas_p=1;
  do
    {
      lucas_p++;
      mpz_set_ui(x,lucas_p); mpz_mul(x,x,x);
      mpz_set_ui(y,4*lucas_q);
      mpz_sub(y,x,y); mpz_mod_ui(y,y,Primzahl);
      mpz_set_ui(x,Primzahl);
    } while (mpz_legendre(y,x)!=-1);
  
  /* nun das Inverse von lucas_p bezglich Primzahl ermitteln */ 
  mpz_set_ui(x,lucas_p); mpz_set_ui(y,Primzahl); mpz_invert(x,x,y);
  lucas_p_inv=mpz_get_ui(x);

  lucas_cache_index=0; /* Cache lschen!! */

  /* jetzt die Wurzel von n errechnen */
  v=lucasv((Primzahl+1)/2);
  /* im folgenden v noch durch 2 teilen (aber modulo Primzahl!!) */
  mpz_set_ui(x,2); mpz_set_ui(y,Primzahl); mpz_invert(x,x,y);
  mpz_set_ui(y,v); mpz_mul(x,x,y);
  v=mpz_mod_ui(y,x,Primzahl);
  return v;
}


void naechste_Primzahl_und_Deltas_ermitteln()
{
  do
    {
      do Primzahl++; while(!is_prime(Primzahl)); /* nchste Primzahl */
      mpz_set_ui(x,Primzahl);
    }
  while (mpz_legendre(kN,x)<0); /* Primzahl gltig fr Primzahlbasis? */
  
  Primzahlen[Primzahl_nr]=Primzahl;
  Primzahl_nr++;
  //printf("%i. Primzahl der Faktorbasis ist %i\r",Primzahl_nr,Primzahl);

  if (Primzahl>Siebgroesse)
    {
      printf("gefundene Primzahl in Faktorbasis zu gro fr das Sieb!\n");
      exit(1);
    }

  if (Primzahl%4==3)
    { /* Lsung kann auf einfache Weise effizient berechnet werden... */
      int d;
      int rm=mpz_mod_ui(x,r,Primzahl);
      mpz_set_ui(x,Primzahl); mpz_powm_ui(y,kN,(Primzahl+1)/4,x);
      d=mpz_get_ui(y); 
      Delta_der_Primzahl[0]=(-rm-d)%Primzahl;
      Delta_der_Primzahl[1]=(-rm+d)%Primzahl;
      //printf("effiziente Deltaermittlung...\n");
    }
  else
    if (Primzahl>100) /* wichtig: keine Lsung durch Lucas-Sequenzen
			 fr Primteiler von kN (und 2)!
			 Auerdem: Fr kleine Zahlen geht Probedivision
			 schneller! */
	{ /* schnelle Lsung mit Lucas-Sequenzen */
	  int d=lucas();
	  int rm=mpz_mod_ui(x,r,Primzahl);
	  Delta_der_Primzahl[0]=(-rm-d)%Primzahl;
	  Delta_der_Primzahl[1]=(-rm+d)%Primzahl;
	  //printf("Deltaermittlung mit Lucassequenzen...\n");
	}
   else
     { /* naive Deltasuche (allgemein, aber langsam fr groe Primzahlen!) */
       int d=0;
       // printf("Naive Deltaermittlung!\n");
       mpz_set_ui(x,Primzahl);
       do
	 {
	   d++; /* "Delta" erhhen */
	   mpz_set(y,r); mpz_add_ui(y,y,d); mpz_mul(y,y,y);
	   mpz_mod(y,y,kN); /* y ist reduziertes Quadrat */
	   mpz_mod(y,y,x); /* teilbar? */
	   /* printf("r+%i mod %i ->  %i\n",d,Primzahl[i],mpz_get_ui(y)); */
	 } while (mpz_get_ui(y)!=0); /* bis teilbar */
       Delta_der_Primzahl[0]=d;
       mpz_mul_ui(y,r,2); mpz_add_ui(y,y,d); mpz_neg(y,y); mpz_mod(y,y,x);
       Delta_der_Primzahl[1]=mpz_get_ui(y);
     }
       
  if (Delta_der_Primzahl[0]<=0) Delta_der_Primzahl[0]+=Primzahl;
  if (Delta_der_Primzahl[1]<=0) Delta_der_Primzahl[1]+=Primzahl;

  /* printf("Primzahl %i ist %i mit Deltas %i,%i\n",
	 Primzahl_nr,Primzahl,Delta_der_Primzahl[0],Delta_der_Primzahl[1]); */
}


void Sieb_initialisieren()
{
 cout << "Sieb von ";
 mpz_out_str(stdout,10,Sieboffset);
 cout << " bis ";
 mpz_add_ui(x,Sieboffset,Siebgroesse); mpz_out_str(stdout,10,x);
 cout << " wird initialisiert." << endl;
 
 mpz_add(x,r,Sieboffset); mpz_mul(x,x,x);
 mpz_sub(x,x,kN);/* x = (trunc(sqrt(kN))+Sieboffset)^2  -kN */

 /* berschlagsmige Initialwerte (geht schneller ;-) ) */
 TSiebElement w = (TSiebElement) (log_Multiplikator*log(fabs(mpz_get_d(x))));
 for (int i=0; i<Siebgroesse; i++) (*Sieb)[i]=w;

 cout << "Initialisierung abgeschlossen." << endl;
}


CGleichungszeile::CGleichungszeile(unsigned int Delta_displacement)
     /* Konstruktor: neue Relation einfgen.
	"Deltas" mit dem Wert Delta initialisieren,
	"Relationen" mit den Primfaktoren des reduzierten Quadrates fllen,
	deren Exponenten ungerade sind. */
{
  //cout << "initialisiere Gleichungszeile" << endl;
  Relation_sparse = new TIntSet;
  Relation_dense=NULL;
  Deltas_sparse= new TIntSet;
  Deltas_dense=NULL;

  mpz_init(Deltas_der_Relationen[gefundene_Relationen]);
  mpz_add_ui(Deltas_der_Relationen[gefundene_Relationen],
	     Sieboffset, Delta_displacement);
  Deltas_sparse->insert(gefundene_Relationen);
  gefundene_Relationen++;
  
  mpz_add(y,r,Deltas_der_Relationen[gefundene_Relationen-1]);
  mpz_mul(y,y,y); mpz_sub(y,y,kN);
  
  int z;
  if (mpz_sgn(y)<0) 
    {
      Relation_sparse->insert(0); 
      /* "nullte Primzahl" ist -1 wegen Bercksichtigung
	 des Vorzeichens. Also die 0 fr Position von  -1 einfgen */
      mpz_neg(y,y); /* -1 war Teiler -> positiv machen */
    }
  
  { /* falls die Routine whrend des Siebens aufgerufen wird, so
       enthlt die Relation die aktuelle Primzahl!
       Deshalb lohnt es sich, diese abzuspalten... */
    z=0;
    while (mpz_div_ui(x,y,Primzahl)==0) { mpz_set(y,x); z++; }
    if (z%2==1) Relation_sparse->insert(Primzahl_nr);
  }
  
  mpz_t wurzel_y;
  mpz_init(wurzel_y); mpz_sqrt(wurzel_y,y);  
  int i;
  for (i=1; i<Primbasis_size && mpz_cmp_ui(wurzel_y,Primzahlen[i])>=0; i++)
    {
      z=0;
      while (mpz_div_ui(x,y,Primzahlen[i])==0)
        { mpz_set(y,x); z++; mpz_sqrt(wurzel_y,y); }
      if (z%2==1)
        Relation_sparse->insert(i); /* ungerade Protenz der i. Primzahl markieren */
    }
  mpz_clear(wurzel_y);

  int p=mpz_get_si(y); /* Ist ein "special-factor" briggeblieben? */
  if (p>1) 
    {
      while (i<Primbasis_size && p!=Primzahlen[i]) i++;
      if (i<Primbasis_size)
	Relation_sparse->insert(i); /* Index der Primzahl eintragen, falls
				       Primzahl in Faktorbasis vorhanden ist */
      else 
	{
	  Relation_sparse->insert(p); /* ansonsten liegt ein Special-Factor
					 auerhalb der Faktorbais vor:
					 ebenfalls einfgen */
	  /* Hinweis: Special-factors werden nicht ber den Index eingefgt,
	     sondern direkt als Zahl. */
	  //cout << "Special-Factor " << p << endl;
	}
    }
  
  if (Relation_sparse->empty()) relevanter_Faktor=no_factor;
  else 
    {
      TIntSet::iterator h = Relation_sparse->end();
      relevanter_Faktor=*(--h); 
      // mit der aufgenommenen Relation lt sich dieser Faktor eliminieren
    }
  
  //cout << "Initialisierung der Gleichungszeile abgeschlossen!" << endl;
}

void CGleichungszeile::Ausgabe()
{
  if (Relation_dense) 
    cout << Relation_dense->count(1) << " Primes (dense) ";
  else copy(Relation_sparse->begin(),Relation_sparse->end(),
	     ostream_iterator<int>(cout," "));
  cout << "<=> ";
  if (Deltas_dense) 
    cout << Deltas_dense->count(1) << " Delta-Werte (dense)";
  else
    cout << Deltas_sparse->size() << " Delta-Werte (sparse)";
  cout << endl;
}

void CGleichungszeile::convert_Relation_to_dense()
{
  //cout << "Konvertiere von Sparsedarstellung nach Densedarstellung." << endl;
  if (Relation_dense)
    { cerr << "Relation ist bereits dense!" << endl; return; }
  Relation_dense = new BitString;
  TIntSet::iterator pos;
  for (pos=Relation_sparse->begin(); pos!=Relation_sparse->end(); ++pos)
    Relation_dense->set(*pos);
  //Relation_sparse->erase(Relation_sparse->begin(),Relation_sparse->end());
  delete Relation_sparse; Relation_sparse=NULL; // Sparsedarstellung lschen
}

void CGleichungszeile::convert_Relation_to_sparse()
{
  //cout << "Konvertiere von Densedarstellung nach Sparsedarstellung." << endl;
  if (Relation_sparse)
    { cerr << "Relation ist bereits sparse!" << endl; return; }
  Relation_sparse = new TIntSet;
  for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
    Relation_sparse->insert(i);
  //Relation_dense->clear();
  delete Relation_dense; Relation_dense=NULL; // Densedarstellung lschen
}

void CGleichungszeile::convert_Deltas_to_dense()
{
  //cout << "Konvertiere von Sparsedarstellung nach Densedarstellung." << endl;
  if (Deltas_dense)
    { cerr << "Deltas ist bereits dense!" << endl; return; }
  Deltas_dense = new BitString;
  TIntSet::iterator pos;
  for (pos=Deltas_sparse->begin(); pos!=Deltas_sparse->end(); ++pos)
    Deltas_dense->set(*pos);
  //Deltas_sparse->erase(Deltas_sparse->begin(),Deltas_sparse->end());
  delete Deltas_sparse; Deltas_sparse=NULL; // Sparsedarstellung lschen
}

void CGleichungszeile::convert_Deltas_to_sparse()
{
  //cout << "Konvertiere von Densedarstellung nach Sparsedarstellung." << endl;
  if (Deltas_sparse)
    { cerr << "Deltas ist bereits sparse!" << endl; return; }
  Deltas_sparse = new TIntSet;
  for (int i=Deltas_dense->first(); i>=0; i=Deltas_dense->next(i))
    Deltas_sparse->insert(i);
  //Deltas_dense->clear();
  delete Deltas_dense; Deltas_dense=NULL; // Densedarstellung lschen
}

void CGleichungszeile::kombinieren(const CGleichungszeile& GL2)
// bergabe von GL2 ber Referenz, um effizienter zu sein...
/* symmetrische Differenzen ermitteln */
{
  //cout << "Kombiniere Gleichungszeilen..." << endl;
  TIntSet::iterator pos;
  /*
    Symmetrische Differenz der Factors in Relation bilden
  */
  if ( GL2.Relation_dense ) /* GL2.Relation is dense */
    if (Relation_dense) /* Relation ist dense */
      (*Relation_dense)^=*GL2.Relation_dense; // symmetrische Differenz bilden
    else /* Relation is sparse */
      {
	convert_Relation_to_dense(); // Relation in dichte Darstellung bringen
	(*Relation_dense)^=*GL2.Relation_dense; /* symmetrische Differenz bilden */
      }
  else /* GL2.Relation is sparse */
    if (Relation_dense) /* and Relation is dense */
      { /* symmetrische Differenz in dense-Darstellung bilden,
	   wobei der andere Vektor sparse ist */
	for (pos=GL2.Relation_sparse->begin(); pos!=GL2.Relation_sparse->end();
	     ++pos) Relation_dense->invert(*pos);
      }
    else
      { /* symmetrische Differenz in sparse-Darstellung bilden */
	Relation_sparse->kombinieren(*GL2.Relation_sparse);
	if (Relation_sparse->size()>Primbasis_size/100)
	  /* Schwelle fr dense berschritten? */
	  {
	    convert_Relation_to_dense(); /* dann von sparse nach dense
					    konvertieren */
	  }
      }
  
  /*
    Symmetrische Differenz der Deltas bilden
  */
  if ( GL2.Deltas_dense ) /* GL2.Deltas is dense */
    if (Deltas_dense) /* Deltas ist dense */
      (*Deltas_dense)^=*GL2.Deltas_dense; /* symmetrische Differenz bilden */
    else /* Deltas is sparse */
      {
	convert_Deltas_to_dense(); /* Deltas in dichte Darstellung bringen */
	(*Deltas_dense)^=*GL2.Deltas_dense; /* symmetrische Differenz bilden */
      }
  else /* GL2.Deltas is sparse */
    if (Deltas_dense) /* and Deltas is dense */
      { /* symmetrische Differenz in dense-Darstellung bilden,
	   wobei der andere Vektor sparse ist */
	for (pos=GL2.Deltas_sparse->begin(); pos!=GL2.Deltas_sparse->end();
	     ++pos) Deltas_dense->invert(*pos);
      }
    else
       { /* symmetrische Differenz in sparse-Darstellung bilden */
	 Deltas_sparse->kombinieren(*GL2.Deltas_sparse);
	 if (Deltas_sparse->size()>Primbasis_size/100)
	   /* Schwelle fr dense berschritten? */
	   {
	     convert_Deltas_to_dense(); /* dann von sparse nach dense
					   konvertieren */
	   }
       }
  
  relevanter_Faktor=groesster_Faktor_in_Relation();
}


TZeilenmenge GLS; /* Gleichungssystem der Faktorbasis */


void Quadratische_Kongruenz_ausrechnen(const TIntSet& Deltas)
{
  /* Alle erzeugten Relationen erfllen die Kongruenz (A=B) mod n.
     Fr das Produkt der zu den in der Menge Deltas gehrigen Relationen
     knnen wir zustzlich die Zahlen A und B konkret ausrechnen.
     Wegen A-B=(A+B)*(A-B)=0 mod n knnen wir daher nun durch ggT(A+B,n)
     einen Teiler von n ermitteln und haben gute Chancen, da es sich dabei
     auch um einen echten (also nichttrivialen) Teiler handelt... */

  cout << "Geeignete Relation ermittelt..." << endl;

  mpz_t A; /* A ist das Produkt der nichtquadrierten Zahlen (mod n),
	      die die Relationen erzeugen */
  mpz_t B; /* B ist die Wurzel des Produkts der erzeugten Relationen,
	      die jeweils in kleine Teiler zerfallen (das war der Grund,
	      warum mit kleinen Primzahlen gesiebt wurde); dieses lt sich
	      nunmehr einfach ermitteln, da wir fr diese Kombination
	      nun wissen, da jeder Faktor gerade-oft auftritt (das war ja
	      der Grund, das Gleichungssystem zu lsen). */
  mpz_t h;
  mpz_init(A); mpz_init(B); mpz_init(h);
  mpz_set_ui(B,1);

  TIntSet::const_iterator pos = Deltas.begin();
  mpz_add(y,r,Deltas_der_Relationen[*pos]);/* Kongruenz des ersten Deltas ermitteln */ 

  mpz_set(A,y); mpz_mod(A,A,kN);
  mpz_mul(x,y,y); mpz_sub(x,x,kN); /* erste Relation ermitteln */
  if (mpz_sgn(x)<0) mpz_neg(x,x); /* -1 ggf. eliminieren */

  while (++pos!=Deltas.end())
    {
      mpz_add(y,r,Deltas_der_Relationen[*pos]); /* Kongruenz des nchsten Deltas ermitteln */ 
      
      mpz_mul(A,A,y); mpz_mod(A,A,kN); /* ausmultiplizieren modulo kN */
      mpz_mul(y,y,y); mpz_sub(y,y,kN); /* und zugehrige Relation ermitteln */
      if (mpz_sgn(y)<0) mpz_neg(y,y); /* -1 ggf. eliminieren */
      mpz_gcd(h,x,y);
      mpz_div(x,x,h); mpz_div(y,y,h); mpz_mul(x,x,y);
      mpz_mul(B,B,h); mpz_mod(B,B,kN);
    }
  if (!mpz_perfect_square_p(x)) cout << "Fehler! x ist kein Quadrat!" << endl;
  mpz_sqrt(x,x); mpz_mul(B,B,x); mpz_mod(B,B,kN);
  
  mpz_out_str(stdout,10,A); printf(" = ");
  mpz_out_str(stdout,10,B); printf(" (mod kN)\n");
   
  mpz_powm_ui(x,A,2,kN); mpz_powm_ui(y,B,2,kN);
  printf("d.h.: ");
  mpz_out_str(stdout,10,x); printf(" = ");
  mpz_out_str(stdout,10,y); printf(" (mod kN)\n");

  mpz_add(x,A,B); mpz_gcd(x,x,n); mpz_div(y,n,x); /* beachte: n und nicht kN
						     ist die eigentlich zu
						     faktorisierende Zahl! */
  mpz_clear(A); mpz_clear(B); mpz_clear(h);

  if ((mpz_cmp_ui(x,1)==0) || (mpz_cmp_ui(y,1)==0))
    {
      cout << "Faktorisierung war leider trivial..." << endl;
    }
  else
    {
      cout << endl << "Faktorisierung gefunden!" << endl;
      mpz_out_str(stdout,10,x); printf(" (%i) * ",mpz_sizeinbase(x,10));
      mpz_out_str(stdout,10,y); printf(" (%i)\n",mpz_sizeinbase(y,10));
      printf("= ");
      mpz_out_str(stdout,10,n); printf(" (%i)\n",mpz_sizeinbase(n,10));
      
      if (!mpz_probab_prime_p(x,25))
	{
	  mpz_out_str(stdout,10,x);
	  printf(" (%i) ist nicht prim!\n",mpz_sizeinbase(x,10));
	}
      if (!mpz_probab_prime_p(y,25))
	{
	  mpz_out_str(stdout,10,y);
	  printf(" (%i) ist nicht prim!\n",mpz_sizeinbase(y,10));
	}
      exit(0); // Programm erfolgreich beenden...
    }
}


void Relation_einfuegen(CGleichungszeile GL)
{ /* hier wird eine neu gefundene Relation in das bisherige Gleichungssystem
     eingefgt. Falls sie redundant ist bzw. zu einer Faktorisierung fhrt,
     so wird dies automatisch erkannt. */
     
  while (!GL.Relation_ist_leer())
    {
      TZeilenmenge::iterator pos = GLS.find(GL);
      if ( pos!=GLS.end())
	{ /* Relation mit diesem Eintrag existiert schon
	     -> neue Kombination mglich! */
	  GL.kombinieren(*pos);       
	}
      else break;
    }
  
  if (GL.Relation_ist_leer()) /* Relation redundant
				 -> Faktorisierung probieren! */
    {
      GL.Ausgabe();
      if (GL.Deltas_dense) GL.convert_Deltas_to_sparse();
      Quadratische_Kongruenz_ausrechnen(*GL.Deltas_sparse);
      GL.clear(); // Gleichungszeile leeren
    }
  else
    {
      GLS.insert(GL);
      
      /* ggf. noch einige Daten ausgeben */ 
      //cout << "Rekonstruktionsschritte: " << Rekonstruktionsschritte << endl;
      //cout << "Gre GLS: " << GLS.size() << endl; 
    }
}


void langsames_aussieben_und_auslesen()
{
 int d,lauf;
 TSiebElement log_Primzahl = (TSiebElement) (log_Multiplikator*log(Primzahl));
 mpz_t laufwert; mpz_init(laufwert);

 // printf("Quadratisches Sieb mit %i\n",Primzahl);

 for ( lauf= (Delta_der_Primzahl[0]!=Delta_der_Primzahl[1]) ? 0 : 1;
       lauf<=1;lauf++)
   {
     //printf("Lauf %i\n",lauf);
     
     mpz_sub_ui(x,Sieboffset,Delta_der_Primzahl[lauf]);
     mpz_mod_ui(x,x,Primzahl);
     d=Primzahl-mpz_get_ui(x);
     
     mpz_set(laufwert,Sieboffset);

     mpz_add_ui(laufwert,laufwert,d);
     mpz_add(laufwert,laufwert,r);

     while (d<Siebgroesse)
       {
         mpz_mul(y,laufwert,laufwert); mpz_sub(y,y,kN);
         //int z=0;
         
         /* if (mpz_mod_ui(x,y,Primzahl)!=0)
           { printf("WARNUNG: Delta-FEHLER!\n"); exit(1); } */
         
         while (mpz_div_ui(y,y,Primzahl)==0)
           {
             //z++;
             (*Sieb)[d]-=log_Primzahl; /* Teiler im Sieb vermerken */
           }
         //if (z>2) printf("%i mit %i: %i mal\n",d,Primzahl,z);
         if ((*Sieb)[d]<2*log_Multiplikator)
           {
             Relation_einfuegen(CGleichungszeile(d));
             /* und Relation vermerken */
           }
         d+=Primzahl; mpz_add_ui(laufwert,laufwert,Primzahl);
       }
   }
 mpz_clear(laufwert);
}



void schnelles_aussieben()
{
 int d,lauf;
 TSiebElement log_Primzahl = (TSiebElement) (log_Multiplikator*log(Primzahl));

 //printf("Quadratisches Sieb mit %i\n",Primzahl);

 for ( lauf= (Delta_der_Primzahl[0]!=Delta_der_Primzahl[1]) ? 0 : 1;
       lauf<=1;lauf++)
   {
     //printf("Lauf %i\n",lauf);
     
     mpz_sub_ui(x,Sieboffset,Delta_der_Primzahl[lauf]);
     mpz_mod_ui(x,x,Primzahl);
     d=Primzahl-mpz_get_ui(x);
     
     while (d<Siebgroesse)
       {
	 (*Sieb)[d]-=log_Primzahl; /* Teiler im Sieb vermerken */
	 d+=Primzahl;
       }   
   }
}


void schnelles_auslesen_des_Siebs()
{
  cout << "Auslesen des (gesiebten) Siebintervalls" << endl;
   TSiebElement log_Threshold = (TSiebElement) (log_Multiplikator *
     log(Factor_Threshold*biggest_Prime_in_Factorbase));
  for (int d=0; d<Siebgroesse; d++)
    if ((*Sieb)[d]<log_Threshold) /* Relation gefunden? */
      Relation_einfuegen(CGleichungszeile(d)); /* dann einfgen */
}


void bestes_kN_ermitteln()
{
  // es ist gnstig, mglichst viele kleine Primfaktoren in der Faktorbasis
  // zu haben. Hierfr kann man die zu faktorisierende Zahl mit einem kleinen
  // Vorfaktor (der am besten prim sein sollte) multiplizieren, um andere
  // Faktorbasen zu erhalten..
  // Welches der gnstigste Vorfaktor ist, wird hier ermittelt.
  // Da der Vorfaktor anschlieend nicht mehr bentigt wird, werden an dieser
  // Stelle die relavanten Werte kN = Vorfaktor*n sowie r = trunc(sqrt(kN))
  // berechnet.
  unsigned int k[] = {1,2,3,5,7,11,13,17,19,0};
  double bestw = 1e20; int best = 0;
  for (int i=0; k[i]!=0; i++)
    {
      mpz_mul_ui(kN,n,k[i]); /* kN initialisieren */
      mpz_sqrt(r,kN); /* Quadratwurzel von kN ermitteln */
      
      double aktw = 0.0;
      unsigned int Primzahl = 1; 
      for (int z=0; z<100; z++)
	{
	  do
	    {
	      do Primzahl++; while(!is_prime(Primzahl)); /* nchste Primzahl */
	      mpz_set_ui(x,Primzahl);
	    }
	  while (mpz_legendre(kN,x)!=1); // Primzahl gltig fr Primzahlbasis?
	  aktw+=log(Primzahl);
	}
      if (aktw<bestw) { bestw=aktw; best=i; } // besserer Multiplikator...
    }
  cout << "Multiplikator fr Sieb ist " << k[best] << endl;
  mpz_mul_ui(kN,n,k[best]); /* kN initialisieren */
  mpz_sqrt(r,kN); /* Quadratwurzel von kN ermitteln */
  mpz_out_str(stdout, 10, r); printf (" ist trunc(sqrt(kN))\n");
}


int main(int argc, char* argv[])
{
  cout << " --------------------------------------" << endl;
  cout << "-     Faktorisierungsprogramm V1.0      -" << endl;
  cout << "-  written 03/1998 by Thorsten Reinecke -" << endl;
  cout << "-     (reinecke@fmi.uni-passau.de)      -" << endl;
  cout << " --------------------------------------" << endl;
  cout << endl;
  cout << "This program may be freely distributed under the" << endl;
  cout << "terms of the GNU-Public-Licence." << endl;
  cout << endl;
    
  if (argc!=2)
  {
    cerr << "Zu faktorisierende Zahl als Parameter erwartet!" << endl;
    exit(0);
  }

  mpz_init(n); /* Variable fr zu faktorisierende Zahl bereitstellen */

  if (mpz_set_str(n, argv[1], 10)) 
  {
    printf("Falscher Parameter: %s\n", argv[1]);
    exit(0);
  }

  mpz_init(kN);
  mpz_init(r);
  mpz_init(x);
  mpz_init(y);

  easy_factor(); /* "einfache" Faktoren rausstreichen */

  printf("Starte Faktorisierung mit Quadratischem Sieb fr\n");
  mpz_out_str(stdout,10,n); printf(" (%i)\n",mpz_sizeinbase(n,10));

  bestes_kN_ermitteln(); /* optimaler Wert, fr den gesiebt werden kann */

  if (mpz_cmp_ui(r,Siebgroesse)<0)
    {
      printf("Sieb ist berdimensioniert und wird verkleinert!\n");
      Siebgroesse=mpz_get_ui(r);
    }

  {
    cout << "Ermittle die Faktorbasis." << endl;
    Primzahlen[0]=-1; /* nullte "Primzahl" ist -1 (Bercksichtigung des
			 Vorzeichens bei den Relationen! */
    Primzahl_nr=1; Primzahl=1;
    while (Primzahl_nr<Primbasis_size)
      {
	naechste_Primzahl_und_Deltas_ermitteln();
	Delta_der_Primzahlen[Primzahl_nr-1]=Delta_der_Primzahl;
      }  
    biggest_Prime_in_Factorbase=Primzahl;
  }
  
  Sieb=new TSieb; /* Speicher fr das Sieb allokieren */
  Sieb->reserve(Siebgroesse); /* Speicherplatz fr Sieb reservieren */
  mpz_init_set_si(Sieboffset,-MAX_Siebgroesse/2); /* Offset initialisieren,
						     bei dem das jeweilige
						     Siebintervall startet */

  /* Quadratischen Reste mit Primzahlen aus der Primzahlbasis aussieben */
  do /* solange Relationen aussieben, bis mehr Relationen gefunden sind
	als Primzahlen in der Primzahlbasis sind (oder eine Faktorisierung
	gefunden ist). */
    {
      Primzahl_nr=1; Primzahl=1;
      Sieb_initialisieren();

      /* wir sieben erst ab einer greren Primzahl aus der Faktorbasis,
	 da wir sowieso unsauber und schnell sieben...
         (je kleiner die Primzahl, um langsamer ist das Sieben mit dieser
	  Zahl; und bei der gewhlten Implementierung ohne Bercksichtigung
	  von Primzahlpotenzen ist auch der zustzlich gemachte Fehler
	  vernachlssigbar und wird zudem ber Factor_Threshold kompensiert.
	  Aber Vorsicht: Startindex nicht zu hoch ansetzen, sonst fallen
	  zu viele Relationen raus!)
      */
      for (Primzahl_nr=5; Primzahl_nr<Primbasis_size; Primzahl_nr++)
	{
	  Primzahl=Primzahlen[Primzahl_nr];
	  Delta_der_Primzahl=Delta_der_Primzahlen[Primzahl_nr];
	  schnelles_aussieben();
	  //langsames_aussieben_und_auslesen(); // alternativ & lahm!
	}
      schnelles_auslesen_des_Siebs();

      cout << "ca. "
	   << 100.0*gefundene_Relationen/(Factor_Threshold*Primbasis_size)
	   << "% der bentigten Relationen gefunden" << endl;
      
      if (mpz_sgn(Sieboffset)>=0)
	mpz_add_ui(Sieboffset,Sieboffset,Siebgroesse);/* neues Siebintervall */
     
      mpz_neg(Sieboffset,Sieboffset); /* Siebintervall abwechselnd
					 positiv und negativ */      
    } while (gefundene_Relationen<=Factor_Threshold*Primbasis_size);

  
  delete Sieb; /* Sieb lschen, d.h. Speicher wieder freigeben */ 

  cout << "no succes..." << endl;
  return 0;
 
}
